/*
 * Copyright (C) 2013 Kaan Onarlioglu <onarliog@ccs.neu.edu>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Kaan Onarlioglu -- www.onarlioglu.com
 * NEU SecLab      -- seclab.ccs.neu.edu
 *
 * perms.c, ver.2013.10.13
 *
 * Privilege drop & escalation helpers for Run-PrivExec.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/capability.h>

#include "perms.h"
#include "utils.h"

/*
 * Capabilities required to run PrivExec without superuser privileges.
 */
#define CAP_LEN 2
static const cap_value_t cap_list[ CAP_LEN] = { CAP_SYS_CHROOT, CAP_SYS_ADMIN};

/*
 * Temporarily drops superuser privileges and capabilities. Currently works with
 * a binary setuid to root, run with sudo, granted file capabilities, or any
 * combination of these.
 *
 * Note that, AFAIK, there is no 100% reliable way of detecting sudo, and there
 * is some guesswork involved here. When setting up the system I would go with
 * either of the setuid or capability options instead.
 */
void drop_privileges() {

    uid_t uid;
    gid_t gid;
    cap_t caps;

    if ( getenv( "SUDO_UID")) { /* sudo */
        uid = atoi( getenv( "SUDO_UID"));
        gid = atoi( getenv( "SUDO_GID"));
    }
    else { /* setuid root */
        uid = getuid();
        gid = getgid();
    }

    setegid( gid);
    seteuid( uid);

    /*
     * ...or you are really running as root, and this whole uid juggling is done
     * in vain. Not my problem.
     */

    /* And disable capabilities, too. */
    caps = cap_get_proc();
    cap_set_flag( caps, CAP_EFFECTIVE, CAP_LEN, cap_list, CAP_CLEAR);
    cap_set_proc( caps);
    cap_free( caps);
}

/*
 * Raises the privileges back to their originals. Again, works with setuid root,
 * sudo, and capabilities.
 */
void raise_privileges() {

    uid_t ruid, euid, suid;
    gid_t rgid, egid, sgid;
    cap_t caps;

    /* setuid and sudo. */
    getresuid( &ruid, &euid, &suid);
    getresgid( &rgid, &egid, &sgid);
    seteuid( suid);
    setegid( sgid);

    /* Capabilities. */
    caps = cap_get_proc();
    cap_set_flag( caps, CAP_EFFECTIVE, CAP_LEN, cap_list, CAP_SET);
    cap_set_proc( caps);
    cap_free( caps);
}

/*
 * This drops sudo for good before executing the private application and tries
 * to erase every trace of sudo and root privileges. HIGHLY EXPERIMENTAL. If
 * possible, don't set up PrivExec for sudo and avoid this altogether.
 */
void drop_sudo_privileges() {

    uid_t uid;
    gid_t gid;
    char *user;
    gid_t *supp;
    int supp_len;

    /* If running sudo... */
    if ( getenv( "SUDO_UID")) {
        uid = atoi( getenv( "SUDO_UID"));
        gid = atoi( getenv( "SUDO_GID"));
        user = getenv( "SUDO_USER");

        /*
         * ...first drop the root's supplemental groups and set the normal
         * user's...
         */
        supp_len = sysconf(_SC_NGROUPS_MAX) + 1;
        supp = malloc( supp_len * sizeof( *supp));
        getgrouplist( user, gid, supp, &supp_len);
        setgroups( supp_len, supp);

        /* ...then correct some environment values... */
        setenv( "HOME", getpwuid( uid)->pw_dir, 1);
        setenv( "USER", user, 1);
        setenv( "USERNAME", user, 1);
        setenv( "LOGNAME", user, 1);

        /* ...and delete others... */
        unsetenv( "SUDO_USER");
        unsetenv( "SUDO_UID");
        unsetenv( "SUDO_GID");
        unsetenv( "SUDO_COMMAND");
        unsetenv( "MAIL");

        /*
         * ...and finally set the group and user IDs. The order is important;
         * first group, then user.
         */
        setregid( gid, gid);
        setreuid( uid, uid);

        free( supp);
    }
}

/* Prints privilege-related information. For debugging. */
#ifdef PRIVEXEC_DEBUG
void print_privileges() {

    uid_t a, s, d;
    gid_t q, w, e;
    cap_t caps;
    char *cap_text;

    /* Print user and group IDs. */
    getresuid( &a, &s, &d);
    getresgid( &q, &w, &e);
    print_green( "UID: %d %d %d\n", a, s, d);
    print_green( "GID: %d %d %d\n", q, w, e);

    /* Print capabilities. */
    caps = cap_get_proc();
    cap_text = cap_to_text( caps, NULL);
    print_green( "CAP: %s\n", cap_text);
    cap_free( caps);
    cap_free( cap_text);
}
#endif /* PRIVEXEC_DEBUG */
